/****************************************************************************
** glue.c
**
** This program joins AGI resources together to form VOL files. It supports
** both AGIv2 and AGIv3 but the user has to tell the program which one. The
** resource files must be set out in the directory structure created by the
** VOLX and XV3 utilities.
**
** (c) Lance Ewing, 1997.
****************************************************************************/

#include <stdio.h>
#include <string.h>
#include <unistd.h>

typedef struct {
   int exists;
   int volNum;
   int length;
   int pos;
} resDirType;

resDirType logdir[256], picdir[256], viewdir[256], snddir[256];
int logDirLen, picDirLen, viewDirLen, sndDirLen;
int lastLogic, lastPicture, lastView, lastSound;

FILE *currentVolFile;
int currentVolNum, currentVolLen;
int agiVer=2, headerLen=5;

unsigned char sierraBuf[20000], outputBuf[20000], sig[9]="";

long getLength(FILE *file)
{
	long tmp;

	fseek(file, 0L, SEEK_END);
	tmp = ftell(file);
	fseek(file, 0L, SEEK_SET);

	return(tmp);
}

/***************************************************************************
** glueResources
**
** Purpose: To glue the resources of one type together into VOL file(s).
***************************************************************************/
void glueResources(char *resType, resDirType *resDir, int *dirLen, int *lastRes)
{
   FILE *resFile;
   int resNum, fileLen, compLen, allSelected=0, doConversion=0;
   unsigned char resFileName[13], volFileName[13], answer[80];
   long startPos;

   printf("Gluing %s resources... ", resType);

   for (resNum=0; resNum<256; resNum++) {
      sprintf(resFileName, "%s\\%s.%d", resType, resType, resNum);
      if ((resFile = fopen(resFileName, "rb")) == NULL) {
         /* Resource is not found */
         resDir[resNum].exists = 0;
      }
      else {
         /* Resource is found */
         *lastRes = resNum;
         resDir[resNum].exists = 1;
         fileLen = getLength(resFile);
         fread(sierraBuf, sizeof(char), fileLen, resFile);
         fclose(resFile);

         /* See if the current VOL file has been filled yet. If it
         ** has, then close it and open the next one. */
         if (currentVolLen + fileLen > 0x0FFFFF) {
            fclose(currentVolFile);
            currentVolNum++;
            sprintf(volFileName, "%sVOL.%d", sig, currentVolNum);
            currentVolFile = fopen(volFileName, "wb");
            currentVolLen = 0;
         }

         startPos = ftell(currentVolFile);

         /* Output the header and compressed data to the VOL file */
         fputc(0x12, currentVolFile);
         fputc(0x34, currentVolFile);
         fputc(currentVolNum, currentVolFile);
         fputc(fileLen & 0xff, currentVolFile);
         fputc((fileLen & 0xff00) >> 8, currentVolFile);
         if (agiVer == 3) {
            fputc(fileLen & 0xff, currentVolFile);
            fputc((fileLen & 0xff00) >> 8, currentVolFile);
         }
         fwrite(sierraBuf, sizeof(char), fileLen, currentVolFile);
         resDir[resNum].volNum = currentVolNum;
         resDir[resNum].length = (fileLen + headerLen);
         resDir[resNum].pos = startPos;
         currentVolLen += (fileLen + headerLen);
      }
   }

   *dirLen = (*lastRes + 1) * 3;
   printf("Done.\n");
}

void writeDirHeader(FILE *dirFile)
{
   int offset;

   offset = 0x0008;
   fputc(0x08, dirFile);
   fputc(0x00, dirFile);
   offset += logDirLen;
   fputc(offset & 0xff, dirFile);
   fputc((offset & 0xff00) >> 8, dirFile);
   offset += picDirLen;
   fputc(offset & 0xff, dirFile);
   fputc((offset & 0xff00) >> 8, dirFile);
   offset += viewDirLen;
   fputc(offset & 0xff, dirFile);
   fputc((offset & 0xff00) >> 8, dirFile);
}

void writeDir3(FILE *dirFile, resDirType *resDir, int lastRes)
{
   int resNum;
   int ch;

   for (resNum=0; resNum<=lastRes; resNum++) {
      if (resDir[resNum].exists) {
         fputc(((resDir[resNum].volNum << 4) & 0xF0) |
               ((resDir[resNum].pos & 0x0F0000) >> 16), dirFile);
         fputc((resDir[resNum].pos & 0x00FF00) >> 8, dirFile);
         fputc((resDir[resNum].pos & 0x0000FF), dirFile);
      }
      else {
         fputc(0xff, dirFile);
         fputc(0xff, dirFile);
         fputc(0xff, dirFile);
      }
   }
}

void buildVolFile()
{
   char volFileName[13];

   currentVolNum = currentVolLen = 0;
   sprintf(volFileName, "%sVOL.0", sig);
   currentVolFile = fopen(volFileName, "wb");
   glueResources("LOGIC", logdir, &logDirLen, &lastLogic);
   glueResources("PICTURE", picdir, &picDirLen, &lastPicture);
   glueResources("VIEW", viewdir, &viewDirLen, &lastView);
   glueResources("SOUND", snddir, &sndDirLen, &lastSound);
   fclose(currentVolFile);
}

void glueGame3()
{
   FILE *glueDirFile;
   char dirFileName[13];

   buildVolFile();
   sprintf(dirFileName, "%sDIR", sig);
   glueDirFile = fopen(dirFileName, "wb");
   writeDirHeader(glueDirFile);
   writeDir3(glueDirFile, logdir, lastLogic);
   writeDir3(glueDirFile, picdir, lastPicture);
   writeDir3(glueDirFile, viewdir, lastView);
   writeDir3(glueDirFile, snddir, lastSound);
   fclose(glueDirFile);
}

void writeDir2(char *dirFileName, resDirType *resDir, int lastRes)
{
   FILE *dirFile;
   int resNum;
   int ch;

   if ((dirFile = fopen(dirFileName, "wb")) == NULL) {
      printf("Unable to create dir file : %s\n", dirFileName);
      exit(1);
   }

   for (resNum=0; resNum<=lastRes; resNum++) {
      if (resDir[resNum].exists) {
         fputc(((resDir[resNum].volNum << 4) & 0xF0) |
               ((resDir[resNum].pos & 0x0F0000) >> 16), dirFile);
         fputc((resDir[resNum].pos & 0x00FF00) >> 8, dirFile);
         fputc((resDir[resNum].pos & 0x0000FF), dirFile);
      }
      else {
         fputc(0xff, dirFile);
         fputc(0xff, dirFile);
         fputc(0xff, dirFile);
      }
   }

   fclose(dirFile);
}

void glueGame2()
{
   buildVolFile();
   writeDir2("LOGDIR", logdir, lastLogic);
   writeDir2("PICDIR", picdir, lastPicture);
   writeDir2("VIEWDIR", viewdir, lastView);
   writeDir2("SNDDIR", snddir, lastSound);
}

void displayHelp()
{
   printf("Usage:  glue [options]\n\n");
   printf("  -2        Create AGIv2 VOL and DIR files (default)\n");
   printf("  -3        Create AGIv3 VOL and DIR files\n");
   printf("  -s sig    Specify the game signature for AGIv3 files. (e.g. SQ2)\n");
   printf("  -h        Display this help message.\n");
   exit(1);
}

void checkDirectories()
{
   int exists;

   exists = chdir("LOGIC");
   if (exists == 0) { chdir(".."); return; }
   exists = chdir("PICTURE");
   if (exists == 0) { chdir(".."); return; }
   exists = chdir("VIEW");
   if (exists == 0) { chdir(".."); return; }
   exists = chdir("SOUND");
   if (exists == 0) { chdir(".."); return; }

   printf("Could not locate the directories containing the resource files.\n");
   printf("(i.e. LOGIC\\, PICTURE\\, VIEW\\, and SOUND\\)\n");
   exit(1);
}

/***************************************************************************
**                              MAIN PROGRAM
***************************************************************************/
void main(int argc, char **argv)
{
   int opt;

   printf("GLUE v1.0    (By Lance Ewing)\n\n");

   if (argc > 1) {
      for (opt=1; opt!=argc; opt++) {
         if (argv[opt][0] == '-') {
            switch(argv[opt][1]) {
               case 'h': displayHelp();
               case '2': agiVer=2; headerLen=5; break;
               case '3': agiVer=3; headerLen=7; break;
               case 's':
                  if (argc > (opt + 1)) {
                     opt++;
                     strcpy(sig, argv[opt]);
                  }
                  else {
                     printf("No signature given for -s option.\n");
                     displayHelp();
                  }
                  break;
	            default:
	               printf("Illegal option : %s\n", argv[opt]);
                  displayHelp();
	         }
	      }
	      else {
	         printf("Illegal option : %s\n", argv[opt]);
            displayHelp();
         }
      }
   }

   checkDirectories();

   switch (agiVer) {
      case 2: glueGame2(); break;
      case 3: glueGame3(); break;
   }
}

